#include "sapporo.h"
#include <sys/time.h>
#include <algorithm>

sapporo_multi_struct sapporo_multi_data[MAXCUDADEVICES];

double get_time() {
  struct timeval Tvalue;
  struct timezone dummy;
  
  gettimeofday(&Tvalue,&dummy);
  return ((double) Tvalue.tv_sec +
	  1.e-6*((double) Tvalue.tv_usec));
}

int sapporo::open(int cluster_id) {
  nCUDAdevices = get_device_count();
  fprintf(stderr, "sapporo::open --- ver 1.5 --- \n");
  if (nCUDAdevices > 0) {
    fprintf(stderr, "sapporo::open - found %d CUDA device(s)\n",
	    nCUDAdevices);
  } else {
    fprintf(stderr, "sapporo::open - FATAL! No CUDA device found\n");
    exit(-1);
  }
  
  nj_max = 131072*nCUDAdevices;
  int which_devices[256];
  for (int i = 0; i < 256; i++) 
    which_devices[i] = i;

  FILE* fd;
  if ((fd = fopen("sapporo.config", "r"))) {
    char line[256];
    fprintf(stderr, "sapporo::open - config file is found\n");
    fgets(line, 256, fd);
    sscanf(line, "%d", &nj_max);
    fgets(line, 256, fd);
    int how_many;
    sscanf(line, "%d", &how_many);
    fprintf(stderr,"                nj_max= %d\n", nj_max);
    fprintf(stderr,"                Request to use %d CUDA device(s): ", how_many);
    if (how_many > nCUDAdevices) {
      fprintf(stderr, " only %d is available. I will use them all.",  nCUDAdevices);
      how_many = nCUDAdevices;
    } else {
      for (int i = 0; i < how_many; i++) {
	fgets(line, 256, fd);
	sscanf(line, "%d", &which_devices[i]);
	fprintf(stderr, " %d ", which_devices[i]);
      }
    }
    fprintf(stderr, "\n");
    nCUDAdevices = how_many;
  } else {
    fprintf(stderr," sapporo::open - no config file is found \n");
    fprintf(stderr,"  using all %d CUDA device(s), nj_max= %d\n", nCUDAdevices, nj_max);
  }

  if (nCUDAdevices > MAXCUDADEVICES) {
    fprintf(stderr, "sapporo::open - WARNING! too many CUDA devices found.\n");
    fprintf(stderr, "                Will use only first %d devices\n", MAXCUDADEVICES);
    nCUDAdevices = MAXCUDADEVICES;
  }


  for (int i = 0; i < nCUDAdevices; i++)
    GPU[i] = new GPUWorker(which_devices[i]);

  for (int i = 0; i < nCUDAdevices; i++) {
    allocate_cuda_memory(i);
  }
  return 0;
}

int sapporo::close(int cluster_id) {
  for (int i = 0; i < nCUDAdevices; i++) {
    free_cuda_memory(i);
  }
  for (int i = 0; i < nCUDAdevices; i++) {
    delete GPU[i];
  }
  return 0;
}

int sapporo::set_ti(int cluster_id, double ti) {
  t_i = to_DS(ti);
  for (int i = 0; i < nCUDAdevices; i++)
    predict[i] = true;
  return 0;
}

int sapporo::get_n_pipes() {
  return n_pipes;
}

int sapporo::set_j_particle(int cluster_id,
			   int address,
			   int id,
			   double tj, double dtj,
			   double mass,
			   double k18[3], double j6[3],
			   double a2[3], double v[3], double x[3]) {

  if (address > nj_max) {
    fprintf(stderr, "Fatal! address= %d > nj_max= %d. I am giving up.\n",
	    address, nj_max);
    exit(-1);
  }

  address_j.push_back(address);

  t_j.push_back( (DS2){to_DS(tj), to_DS(dtj)} );
  
  DS  Dmass = (DS){mass, INT_AS_FLOAT(id)};
  pos_j.push_back( (DS4){to_DS(x[0]), to_DS(x[1]), to_DS(x[2]), Dmass} );
  vel_j.push_back( (float4){v[0],    v[1],    v[2],    0.0} );
  acc_j.push_back( (float4){a2[0]*2, a2[1]*2, a2[2]*2, 0.0} );
  jrk_j.push_back( (float4){j6[0]*6, j6[1]*6, j6[2]*6, 0.0} );
  for (int i = 0; i < nCUDAdevices; i++)
    nj_modified[i] = address_j.size();

  return 0;
};

void sapporo::calc_firsthalf(int cluster_id,
			    int nj, int ni,
			    int id[], 
			    double xi[][3], double vi[][3],
			    double aold[][3], double j6old[][3],
			    double phiold[3], 
			    double eps2, double h2[]) {
  for (int i = 0; i < ni; i++) {
    DS Dmass = (DS){h2[i], INT_AS_FLOAT(id[i])};
    pos_i[i] = (DS4) { to_DS(xi[i][0]),
		       to_DS(xi[i][1]),
		       to_DS(xi[i][2]),
		       Dmass };
    vel_i[i] = (float4){ vi[i][0], vi[i][1], vi[i][2], eps2};
    EPS2 = eps2;

  }

  if (address_j.size() > 0) {
    for (int i = 0; i < nCUDAdevices; i++) {
      send_j_particles_to_device(i);
    }
  }

  for (int i = 0; i < nCUDAdevices; i++) {
    send_i_particles_to_device(i, ni);
  }
  
  for (int i = 0; i < nCUDAdevices; i++)
    GPU[i]->sync();
  
  if(address_j.size() > 0) {
    address_j.clear();
    t_j.clear();
    pos_j.clear();
    vel_j.clear();
    acc_j.clear();
    jrk_j.clear();
  }
  evaluate_gravity(ni, nj);

}

int sapporo::calc_lasthalf(int cluster_id,
			  int nj, int ni,
			  int index[], 
			  double xi[][3], double vi[][3],
			  double eps2, double h2[],
			  double acc[][3], double jerk[][3], double pot[]) {

  for (int i = 0; i < ni; i++) {
    pot[i] = 0;
     acc[i][0] =  acc[i][1] =  acc[i][2] = 0;
    jerk[i][0] = jerk[i][1] = jerk[i][2] = 0;
  }
  for (int dev = 0; dev < nCUDAdevices; dev++) {
    fetch_data_from_device(dev, ni);
    GPU[dev]->sync(); 
 
    for (int i = 0; i < ni; i++) {
      float4 acci = acc_i[i];
      float4 jrki = jrk_i[i];
      pot[i]    += acci.w;
      
      acc[i][0] += acci.x;
      acc[i][1] += acci.y;
      acc[i][2] += acci.z;
      
      jerk[i][0] += jrki.x;
      jerk[i][1] += jrki.y;
      jerk[i][2] += jrki.z;
      
    }
  }
  return 0;
};


int sapporo::calc_lasthalf2(int cluster_id,
			   int nj, int ni,
			   int index[], 
			   double xi[][3], double vi[][3],
			   double eps2, double h2[],
			   double acc[][3], double jerk[][3], double pot[],
			   int nnbindex[]) {
  

  float ds_min[NTHREADS];
  for (int i = 0; i < ni; i++) {
    pot[i] = 0;
    acc[i][0]  = acc[i][1]  = acc[i][2]  = 0;
    jerk[i][0] = jerk[i][1] = jerk[i][2] = 0;
    nnbindex[i] = 0;
    ds_min[i] = 1.0e10;
  }

  for (int dev = 0; dev < nCUDAdevices; dev++) {
    double t1 = get_time();
    fetch_data_from_device(dev, ni);
    GPU[dev]->sync();
 
    for (int i = 0; i < ni; i++) {
      
      float4 acci = acc_i[i];
      float4 jrki = jrk_i[i];
      float  ds   = ds_i[i];

//       fprintf(stdout, "device= %d, ni= %d pot = %g\n",
// 	      dev, i, acci.x);

      pot[i]    += acci.w;
      
      acc[i][0] += acci.x;
      acc[i][1] += acci.y;
      acc[i][2] += acci.z;
      
      jerk[i][0] += jrki.x;
      jerk[i][1] += jrki.y;
      jerk[i][2] += jrki.z;

      if (ds < ds_min[i]) {
	int nnb = (int)(jrki.w);
	nnbindex[i] = nnb; 
	ds_min[i] = ds;
      }
      
    }
  }
  return 0;
};

int sapporo::read_ngb_list(int cluster_id) {
  bool overflow = false;
  int ni[MAXCUDADEVICES];
  for (int dev = 0; dev < nCUDAdevices; dev++) {
    ni[dev] = fetch_ngb_list_from_device(dev);
  }
  for(int dev = 0; dev < nCUDAdevices; dev++) {
    GPU[dev]->sync(); 
    for (int j = 0; j < ni[dev]; j++)
      if (ngb_list_i[dev*NGB_PP*n_pipes + j*NGB_PP] >= NGB_PP) {
	overflow = true;
      }
  }
    
  return overflow;
}

int sapporo::get_ngb_list(int cluster_id,
			 int ipipe,
			 int maxlength,
			 int &nblen,
			 int nbl[]) {
 
  if (ipipe >= devs[0].ni) {
    fprintf(stderr, "Fatal! ipipe= %d >= dev.ni= %d. I give up.\n",
	    ipipe, devs[0].ni);
    exit(-1);
  }

  bool overflow = false;
  nblen = 0;
  for (int i = 0; i < nCUDAdevices; i++) {
    int offset = i*NGB_PP*n_pipes + NGB_PP*ipipe;
    int len = ngb_list_i[offset];
    memcpy(nbl+nblen, &ngb_list_i[offset+1], sizeof(int)*min(len, maxlength - len));
    nblen += len;
    if (nblen >= maxlength) {
      overflow = true;
      break;
    }
  }
  sort(nbl, nbl + min(nblen, maxlength));

//   if (overflow) {
//     fprintf(stderr, "sapporo::get_ngb_list(..) - overflow for ipipe= %d, ngb= %d\n", 
// 	    ipipe, nblen);
//   }
  
  return overflow;
}
